#include <fstream>
#include <cmath>
#include <cstring>
#include "trimesh.h"
#include "iomesh.h"
#include "hashgrid.h"

// parameters
float radius;
float dx;
const char *particle_file_format;
const char *output_file_format;
int initial_frame;
int final_frame;

// particle data
std::vector<Vec3f> px;
HashGrid3D particle_grid;

// marching tetrahedra blobby data
TriMesh mesh;
std::vector<Vec3f> bx, bnormal;
HashTable<Vec3i, char> blobby_grid;

// functions found below
bool format_is_variable(const char *fmt);
bool read_particles(int frame);
void marching_tets(void);
void contour_cube(int i, int j, int k);
void contour_tet(const Vec3f &x0, Vec3f &x1, Vec3f &x2, Vec3f &x3, float p0, float p1, float p2, float p3);
void make_triangle(const Vec3f &x0, const Vec3f &x1, const Vec3f &x2);
Vec3f edgecross(const Vec3f &x0, const Vec3f &x1, float p0, float p1);
float eval_blobby(const Vec3f &x);
void eval_blobby_normal(const Vec3f &x, Vec3f &normal);

//====================================================================================================

int main(int argc, char **argv)
{
   if(argc!=5){
      std::cerr<<"usage: make_blobbies <particle_filename_format> <output_filename_format> <radius> <dx>"<<std::endl;
      std::exit(1);
   }

   particle_file_format=argv[1];
   output_file_format=argv[2];
   std::sscanf(argv[3], "%f", &radius);
   std::sscanf(argv[4], "%f", &dx);

   particle_grid.set_grid_size(radius);

   std::cout<<"radius=="<<radius<<" and dx=="<<dx<<std::endl;

   if(format_is_variable(particle_file_format)){
      for(int frame=0; read_particles(frame); ++frame){
         marching_tets();
         std::cout<<"  finished contouring frame "<<frame<<std::endl;
         write_objfile(mesh, bx, bnormal, output_file_format, frame);
         std::cout<<"  wrote triangle mesh"<<std::endl;
      }
   }else{
      if(read_particles(0)){
         marching_tets();
         std::cout<<"  finished contouring"<<std::endl;
         write_objfile(mesh, bx, bnormal, output_file_format);
         std::cout<<"  wrote triangle mesh"<<std::endl;
      }
   }

   return 0;
}

//====================================================================================================

bool format_is_variable(const char *fmt)
{
   for(int c=0; fmt[c]!=0; ++c){
      if(fmt[c]=='%' && fmt[c+1]!='%')
	 return true;
   }
   return false;
}

bool read_particles(int frame)
{
   char filename[1024];
   std::sprintf(filename, particle_file_format, frame);
   std::ifstream particle_in(filename);
   if(particle_in.bad() || particle_in.eof())
      return false;

   int np=-1;
   particle_in>>np;
   if(particle_in.bad() || np<=0 || np>10000000){
      std::cerr<<"probably an error in particle file ["<<filename<<"]: number of particles="<<np<<std::endl;
      std::exit(2);
   }

   particle_grid.clear();
   px.resize(np);
   Vec3f v;
   for(int i=0; i<np; ++i){
      particle_in>>px[i];
      particle_in>>v;
      particle_grid.add_point(px[i], i);
   }
   if(particle_in.bad()){
      std::cerr<<"error reading particle file ["<<filename<<"]"<<std::endl;
      std::exit(3);
   }

   std::cout<<"read ["<<filename<<"]"<<std::endl;

   return true;
}

void marching_tets(void)
{
   mesh.clear();
   bx.clear();
   bnormal.clear();
   blobby_grid.clear();
   // we only want to do marching tets near particles, so loop over the particles:
   for(unsigned int p=0; p<px.size(); ++p){
      int imin=(int)std::floor((px[p][0]-radius)/dx), imax=(int)std::floor((px[p][0]+radius)/dx),
          jmin=(int)std::floor((px[p][1]-radius)/dx), jmax=(int)std::floor((px[p][1]+radius)/dx),
          kmin=(int)std::floor((px[p][2]-radius)/dx), kmax=(int)std::floor((px[p][2]+radius)/dx);
      for(int i=imin; i<=imax; ++i) for(int j=jmin; j<=jmax; ++j) for(int k=kmin; k<=kmax; ++k){
         if(!blobby_grid.has_entry(Vec3i(i,j,k))){ // if we haven't done this grid cell before
            blobby_grid.add(Vec3i(i,j,k), 1); // then mark that we're doing it
            contour_cube(i,j,k);
         }
      }
   }
}

void contour_cube(int i, int j, int k)
{
   Vec3f x000(i*dx, j*dx, k*dx),         x001(i*dx, j*dx, (k+1)*dx),
         x010(i*dx, (j+1)*dx, k*dx),     x011(i*dx, (j+1)*dx, (k+1)*dx),
         x100((i+1)*dx, j*dx, k*dx),     x101((i+1)*dx, j*dx, (k+1)*dx),
         x110((i+1)*dx, (j+1)*dx, k*dx), x111((i+1)*dx, (j+1)*dx, (k+1)*dx);
   float p000=eval_blobby(x000), p001=eval_blobby(x001),
         p010=eval_blobby(x010), p011=eval_blobby(x011),
         p100=eval_blobby(x100), p101=eval_blobby(x101),
         p110=eval_blobby(x110), p111=eval_blobby(x111);

   if((i+j+k)%2){
      contour_tet(x000, x110, x011, x010, p000, p110, p011, p010);
      contour_tet(x000, x101, x110, x100, p000, p101, p110, p100);
      contour_tet(x101, x011, x110, x111, p101, p011, p110, p111);
      contour_tet(x000, x011, x101, x001, p000, p011, p101, p001);
      contour_tet(x000, x110, x101, x011, p000, p110, p101, p011);
   }else{
      contour_tet(x100, x010, x001, x000, p100, p010, p001, p000);
      contour_tet(x100, x111, x010, x110, p100, p111, p010, p110);
      contour_tet(x100, x001, x111, x101, p100, p001, p111, p101);
      contour_tet(x001, x010, x111, x011, p001, p010, p111, p011);
      contour_tet(x100, x010, x111, x001, p100, p010, p111, p001);
   }
}

// corners arranged so that 0-1-2 uses right-hand-rule to get to 3
void contour_tet(const Vec3f &x0, Vec3f &x1, Vec3f &x2, Vec3f &x3, float p0, float p1, float p2, float p3)
{
   // guard against topological degeneracies
   if(p0==0) p0=1e-30;
   if(p1==0) p1=1e-30;
   if(p2==0) p2=1e-30;
   if(p3==0) p3=1e-30;

   if(p0<0){
      if(p1<0){
         if(p2<0){
            if(p3<0){
               return; // no contour here
            }else // p3>=0
               make_triangle(edgecross(x0,x3,p0,p3), edgecross(x1,x3,p1,p3), edgecross(x2,x3,p2,p3));
         }else{ // p2>=0
            if(p3<0)
               make_triangle(edgecross(x0,x2,p0,p2), edgecross(x3,x2,p3,p2), edgecross(x1,x2,p1,p2));
            else{ // p3>=0
               make_triangle(edgecross(x0,x3,p0,p3), edgecross(x1,x3,p1,p3), edgecross(x0,x2,p0,p2));
               make_triangle(edgecross(x1,x3,p1,p3), edgecross(x1,x2,p1,p2), edgecross(x0,x2,p0,p2));
            }
         }
      }else{ // p1>=0
         if(p2<0){
            if(p3<0)
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x2,x1,p2,p1), edgecross(x3,x1,p3,p1));
            else{ // p3>=0
               make_triangle(edgecross(x0,x3,p0,p3), edgecross(x0,x1,p0,p1), edgecross(x2,x3,p2,p3));
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x2,x1,p2,p1), edgecross(x2,x3,p2,p3));
            }
         }else{ // p2>=0
            if(p3<0){
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x0,x2,p0,p2), edgecross(x3,x2,p3,p2));
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x3,x2,p3,p2), edgecross(x3,x1,p3,p1));
            }else // p3>=0
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x0,x2,p0,p2), edgecross(x0,x3,p0,p3));
         }
      }
   }else{ // p0>=0
      if(p1<0){
         if(p2<0){
            if(p3<0)
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x0,x3,p0,p3), edgecross(x0,x2,p0,p2));
            else{ // p3>=0
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x3,x1,p3,p1), edgecross(x3,x2,p3,p2));
               make_triangle(edgecross(x3,x2,p3,p2), edgecross(x0,x2,p0,p2), edgecross(x0,x1,p0,p1));
            }
         }else{ // p2>=0
            if(p3<0){
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x0,x3,p0,p3), edgecross(x3,x2,p3,p2));
               make_triangle(edgecross(x0,x1,p0,p1), edgecross(x3,x2,p3,p2), edgecross(x2,x1,p2,p1));
            }else // p3>=0
               make_triangle(edgecross(x1,x0,p1,p0), edgecross(x1,x3,p1,p3), edgecross(x1,x2,p1,p2));
         }
      }else{ // p1>=0
         if(p2<0){
            if(p3<0){
               make_triangle(edgecross(x1,x3,p1,p3), edgecross(x0,x3,p0,p3), edgecross(x0,x2,p0,p2));
               make_triangle(edgecross(x1,x3,p1,p3), edgecross(x0,x2,p0,p2), edgecross(x1,x2,p1,p2));
            }else // p3>=0
               make_triangle(edgecross(x0,x2,p0,p2), edgecross(x1,x2,p1,p2), edgecross(x3,x2,p3,p2));
         }else{ // p2>=0
            if(p3<0)
               make_triangle(edgecross(x0,x3,p0,p3), edgecross(x2,x3,p2,p3), edgecross(x1,x3,p1,p3));
            else{ // p3>=0
               return; // assume no degenerate cases (where some of the p's are zero)
            }
         }
      }
   }
}

void make_triangle(const Vec3f &x0, const Vec3f &x1, const Vec3f &x2)
{
   int n=bx.size();
   mesh.tri.push_back(Vec3i(n,n+1,n+2));
   bx.push_back(x0);
   bx.push_back(x2);
   bx.push_back(x1);
   bnormal.resize(bx.size());
   eval_blobby_normal(x0, bnormal[n]);
   eval_blobby_normal(x2, bnormal[n+1]);
   eval_blobby_normal(x1, bnormal[n+2]);
}

Vec3f edgecross(const Vec3f &x0, const Vec3f &x1, float p0, float p1)
{
   float alpha=p1/(p1-p0);
   return alpha*x0+(1-alpha)*x1;
}

float eval_blobby(const Vec3f &x)
{
   static std::vector<int> nbrs;
   particle_grid.find_box(x-Vec3f(radius), x+Vec3f(radius), nbrs);
   float value=0;
   for(unsigned int i=0; i<nbrs.size(); ++i){
      float d=dist(x,px[nbrs[i]]);
      if(d<radius)
         value+=sqr(radius-d);
   }
   return 0.5*radius-std::sqrt(value); // negative means inside, positive means outside
}

void eval_blobby_normal(const Vec3f &x, Vec3f &normal)
{
   normal[0]=normal[1]=normal[2]=0;

   static std::vector<int> nbrs;
   particle_grid.find_box(x-Vec3f(radius), x+Vec3f(radius), nbrs);
   if(nbrs.empty()){
      std::cerr<<"trying to find a normal in an empty region of space"<<std::endl;
      return;
   }

   for(unsigned int i=0; i<nbrs.size(); ++i){
      float d=dist(x,px[nbrs[i]]);
      if(d<radius){
         if(d<0.01*radius)
            d=0.01*radius;
         normal+=(1/d-1/radius)*(x-px[nbrs[i]]);
      }
   }
   normalize(normal);
}

