#include <fstream>
#include <cmath>
#include <cstring>
#include "trimesh.h"
#include "iomesh.h"
#include "hashgrid.h"

// simulation parameters
float simulation_time; // in seconds
float frame_rate; // in frames per second
Vec3f gravity(0,-9.81,0); // standard value
float mass; // of each particle
float spring_stiffness;
float rest_length;
float interaction_radius;
float spring_damping;
float friction_coefficient;
const char *collision_geometry_filename;
const char *particle_file_format;
int initial_frame;
int final_frame;

// particle data
std::vector<Vec3f> x, xnew, v, f;
HashGrid3D particle_grid;

// collision geometry data
TriMesh collision_mesh;
std::vector<Vec3f> collision_x;
HashGrid3D collision_grid;

// functions found below
void get_configuration(int argc, char **argv);
void read_collision_geometry(void);
void read_initial_particles(void);
void write_particles(int frame);
void advance_frame(void);
float select_timestep(void);
void advance_particle_dynamics(float dt);
void compute_all_forces(void);
void compute_particle_force(int i, int j, Vec3f &fij);
void process_collisions(float dt);
bool check_one_collision(int i, int t);
float orient(const Vec3f &x1, const Vec3f &x2, const Vec3f &x3, const Vec3f &x4);
bool intersect_tri_seg(const Vec3f &x1, const Vec3f &x2, const Vec3f &x3, const Vec3f &x4, const Vec3f &x5);

//====================================================================================================

int main(int argc, char **argv)
{
   std::cout<<"starting up..."<<std::endl;
   get_configuration(argc, argv);
   std::cout<<"read configuration"<<std::endl;
   read_collision_geometry();
   std::cout<<"read collision geometry (#tri="<<collision_mesh.tri.size()<<", #v="<<collision_x.size()<<")"<<std::endl;
   read_initial_particles();
   std::cout<<"read initial particles (#particles="<<x.size()<<")"<<std::endl;

   for(int frame=initial_frame+1; frame<=final_frame; ++frame){
      std::cout<<"---------------------------------------\nadvancing to frame "<<frame<<std::endl;
      advance_frame();
      std::cout<<"writing frame "<<frame<<std::endl;
      write_particles(frame);
   }

   return 0;
}

//====================================================================================================

void get_configuration(int argc, char **argv)
{
   if(argc<4){
      std::cerr<<"usage: simulate_particles <config_file> <collision_geometry> <particle_format> [frame_number]"<<std::endl;
      std::exit(1);
   }
   {
      std::ifstream config_in(argv[1]);
      config_in>>simulation_time;
      config_in>>frame_rate;
      config_in>>mass;
      config_in>>spring_stiffness;
      config_in>>rest_length;
      config_in>>interaction_radius;
      config_in>>spring_damping;
      config_in>>friction_coefficient;
      if(config_in.bad()){
         std::cerr<<"problem reading configuration file ["<<argv[1]<<"]"<<std::endl;
         std::exit(2);
      }
   }
   collision_geometry_filename=argv[2];
   particle_file_format=argv[3];
   if(argc>4)
      std::sscanf(argv[4], "%d", &initial_frame);
   else
      initial_frame=0;
   final_frame=lround(simulation_time*frame_rate);
}

void read_collision_geometry(void)
{
   if(!read_objfile(collision_mesh, collision_x, "%s", collision_geometry_filename)){
      std::cerr<<"Failed to read collision geometry in ["<<collision_geometry_filename<<"]"<<std::endl;
      std::exit(3);
   }

   float average_length=0;
   for(unsigned int t=0; t<collision_mesh.tri.size(); ++t){
      int i,j,k; assign(collision_mesh.tri[t], i, j, k);
      average_length+=dist(collision_x[i], collision_x[j]);
      average_length+=dist(collision_x[j], collision_x[k]);
      average_length+=dist(collision_x[k], collision_x[i]);
   }
   average_length/=3*collision_mesh.tri.size();
   collision_grid.set_grid_size(average_length);
   float safety_margin=1e-3*average_length;

   for(unsigned int t=0; t<collision_mesh.tri.size(); ++t){
      int i,j,k; assign(collision_mesh.tri[t], i, j, k);
      Vec3f xmin(min(collision_x[i][0], collision_x[j][0], collision_x[k][0])-safety_margin,
                 min(collision_x[i][1], collision_x[j][1], collision_x[k][1])-safety_margin,
                 min(collision_x[i][2], collision_x[j][2], collision_x[k][2])-safety_margin),
            xmax(max(collision_x[i][0], collision_x[j][0], collision_x[k][0])+safety_margin,
                 max(collision_x[i][1], collision_x[j][1], collision_x[k][1])+safety_margin,
                 max(collision_x[i][2], collision_x[j][2], collision_x[k][2])+safety_margin);
      collision_grid.add_box(xmin, xmax, t);
   }
}

void read_initial_particles(void)
{
   char filename[1024];
   std::sprintf(filename, particle_file_format, initial_frame);
   std::ifstream particle_in(filename);
   int np=0;
   particle_in>>np;
   if(particle_in.bad() || np<=0 || np>10000000){
      std::cerr<<"probably an error in particle file ["<<filename<<"]: number of particles="<<np<<std::endl;
      std::exit(4);
   }
   x.resize(np);
   xnew.resize(np);
   v.resize(np);
   f.resize(np);
   for(int i=0; i<np; ++i){
      particle_in>>x[i];
      particle_in>>v[i];
   }
   if(particle_in.bad()){
      std::cerr<<"error reading particle file ["<<filename<<"]"<<std::endl;
      std::exit(5);
   }
}

void write_particles(int frame)
{
   char filename[1024];
   std::sprintf(filename, particle_file_format, frame);
   std::ofstream particle_out(filename);
   particle_out<<x.size()<<std::endl;
   for(unsigned int i=0; i<x.size(); ++i){
      particle_out<<x[i]<<'\n'<<v[i]<<std::endl;
   }
   if(particle_out.bad()){
      std::cerr<<"Error writing particles in frame "<<frame<<" to file ["<<filename<<"]"<<std::endl;
      std::exit(6);
   }
}

void advance_frame(void)
{
   float t=0, endt=1/frame_rate;
   bool finished=false;
   while(!finished){
      // first set up the particles' acceleration structure
      particle_grid.clear();
      for(unsigned int i=0; i<x.size(); ++i)
         particle_grid.add_box(x[i]-Vec3f(interaction_radius), x[i]+Vec3f(interaction_radius), i);
      // then figure out a stable time step
      float dt=select_timestep();
      if(t+dt>=endt){
         dt=endt-t;
         finished=true;
      }else if(t+1.5*dt>=endt)
         dt=0.5*(endt-t);
      std::cout<<" substep: "<<dt<<" (from "<<t<<" to get to "<<100*(t+dt)/endt<<"% of frame)"<<std::endl;
      // and then run internal dynamics
      std::cout<<"   computing dynamics..."<<std::endl;
      advance_particle_dynamics(dt);
      // and collision dynamics
      std::cout<<"   processing collisions..."<<std::endl;
      process_collisions(dt);
      // finish up with new positions and new time
      x.swap(xnew);
      t+=dt;
   }
}

float select_timestep(void)
{
   @@@ you will need to determine a reasonable time step size here
}

void advance_particle_dynamics(float dt)
{
   compute_all_forces();
   @@@ you will need to implement one step of Symplectic Euler here
   @@@ to get the new positions in xnew[] and overwrite the new velocities into v[]
   @@@ (leave the old positions in x[] untouched --- the collision processing depends on them)
}

void compute_all_forces(void)
{
   unsigned int i, n=x.size();
   // start with gravity
   for(i=0; i<n; ++i)
      f[i]=mass*gravity;
   // next find inter-particle forces
   std::vector<int> nbrs;
   Vec3f fij;
   for(i=0; i<n; ++i){
      particle_grid.find_point(x[i], nbrs);
      for(unsigned int r=0; r<nbrs.size(); ++r){
         unsigned int j=nbrs[r];
         if(j>i){ // only compute the force for one of the pair of particles
            compute_particle_force(i, j, fij);
            f[i]+=fij;
            f[j]-=fij;
         }
      }
   }
}

void compute_particle_force(int i, int j, Vec3f &fij)
{
   @@@ compute the force on particle i due to particle j and store it in fij
}

void process_collisions(float dt)
{
   unsigned int p, n=x.size();
   std::vector<int> trilist;
   for(p=0; p<n; ++p){
      bool collisions_solved;
      int collision_iteration=0;
      Vec3f xnew_precollision=xnew[p];
      do{
         collisions_solved=true;
         Vec3f xmin(std::min(x[p][0], xnew[p][0]),
                    std::min(x[p][1], xnew[p][1]),
                    std::min(x[p][2], xnew[p][2])),
               xmax(std::max(x[p][0], xnew[p][0]),
                    std::max(x[p][1], xnew[p][1]),
                    std::max(x[p][2], xnew[p][2]));
         collision_grid.find_box(xmin, xmax, trilist);
         for(unsigned int r=0; r<trilist.size(); ++r){
            if(check_one_collision(p, trilist[r])){
               collisions_solved=false; // might now be intersecting a previously checked triangle
               break; // new trajectory, so go and regenerate list of possible triangles
            }
         }
         ++collision_iteration;
         if(!collisions_solved && collision_iteration>10){
            // just give up and use the old (presumably good) position
            xnew[p]=x[p];
            collisions_solved=true;
         }
      }while(!collisions_solved);
      // include the effect of the collision impulse in the velocity
      v[p]+=(xnew[p]-xnew_precollision)/dt;
   }
}

bool check_one_collision(int p, int t)
{
   int i,j,k; assign(collision_mesh.tri[t], i, j, k);
   if(intersect_tri_seg(collision_x[i], collision_x[j], collision_x[k], x[p], xnew[p])){
      @@@ process the inelastic frictional collision between point p and triangle ijk,
      @@@ changing xnew[p] to account for it
      @@@ (note that x[p] and v[p] should be untouched, and in fact v[p] is irrelevant)
      return true;
   }else
      return false;
}

float orient(const Vec3f &x1, const Vec3f &x2, const Vec3f &x3, const Vec3f &x4)
{
   return (x1[0]-x4[0])*( (x2[1]-x4[1])*(x3[2]-x4[2]) - (x3[1]-x4[1])*(x2[2]-x4[2]) )
         +(x2[0]-x4[0])*( (x3[1]-x4[1])*(x1[2]-x4[2]) - (x1[1]-x4[1])*(x3[2]-x4[2]) )
         +(x3[0]-x4[0])*( (x1[1]-x4[1])*(x2[2]-x4[2]) - (x2[1]-x4[1])*(x1[2]-x4[2]) );
}

// check triangle 1-2-3 against segment 4-5
bool intersect_tri_seg(const Vec3f &x1, const Vec3f &x2, const Vec3f &x3, const Vec3f &x4, const Vec3f &x5)
{
   // note: comparing this triangle to another edge sharing vertex 4 or 5 will encounter exactly the same
   // floating-point calculation of a determinant, so this segment-plane test is watertight.
   float d=orient(x1, x2, x3, x5);
   float e=-orient(x1, x2, x3, x4);
   if((d>0) ^ (e>0))
      return false;
   // note: using the triangle edges in the first two spots guarantees the same floating point result (up to sign)
   // if we transpose the triangle vertices -- e.g. testing an adjacent triangle -- so this triangle-line test is
   // watertight.
   float a=orient(x2, x3, x4, x5);
   float b=orient(x3, x1, x4, x5);
   if((a>0) ^ (b>0))
      return false;
   float c=orient(x1, x2, x4, x5);
   if((a>0) ^ (c>0))
      return false;
   if(a+b+c==0.0) // degenerate situation
      return false;
   return true;
}

