#include <cmath>
#include <cstdio>
#include <fstream>
#include "gluvi.h"
#include "trimesh.h"
#include "display.h"
#include "iomesh.h"

int frame;
char frametext[50];
char **filename_formats;
std::vector<TriMesh> mesh;
std::vector<std::vector<Vec3f> > x, normal;

const int numcolours=11;
const float colours[33]={1, 1, 1,
                         1, 1, 0,
                         1, 0, 1,
                         0, 1, 1,
                         1, 0, 0,
                         0, 1, 0,
                         0, 0, 1,
                         0.5, 0.5, 0.5,
                         0, 0.5, 1,
                         1, 0, 0.5,
                         0.5, 1, 0};

void display(void)
{
   Gluvi::set_generic_lights();
   for(unsigned int i=0; i<mesh.size(); ++i){
      Gluvi::set_generic_material(colours[3*((2*i)%numcolours)],
                                  colours[1+3*((2*i)%numcolours)],
                                  colours[2+3*((2*i)%numcolours)], GL_FRONT);
      Gluvi::set_generic_material(colours[3*((2*i+1)%numcolours)],
                                  colours[1+3*((2*i+1)%numcolours)],
                                  colours[2+3*((2*i+1)%numcolours)], GL_BACK);
      display_mesh(mesh[i], x[i], normal[i]);
   }
}

bool read_frame(int newframe)
{
   if(newframe<0)
      return false;

   for(unsigned int i=0; i<mesh.size(); ++i){
      if(!read_objfile(mesh[i], x[i], normal[i], filename_formats[i], newframe)){
         // then go back and read the old frame
         for(unsigned int j=0; j<=i; ++j)
            read_objfile(mesh[j], x[j], normal[j], filename_formats[j], frame);
         return false;
      }
   }
   frame=newframe;
   std::sprintf(frametext, "frame %04d", frame);
   return true;
}

void set_camera(Gluvi::Target3D &cam)
{
   if(mesh.size()==0)
      return;
   float xmin=MAXFLOAT, xmax=-MAXFLOAT,
         ymin=MAXFLOAT, ymax=-MAXFLOAT,
         zmin=MAXFLOAT, zmax=-MAXFLOAT;
   for(unsigned int i=0; i<x.size(); ++i){
      for(unsigned int j=0; j<x[i].size(); ++j){
         if(x[i][j][0]<xmin) xmin=x[i][j][0];
         if(x[i][j][0]>xmax) xmax=x[i][j][0];
         if(x[i][j][1]<ymin) ymin=x[i][j][1];
         if(x[i][j][1]>ymax) ymax=x[i][j][1];
         if(x[i][j][2]<zmin) zmin=x[i][j][2];
         if(x[i][j][2]>zmax) zmax=x[i][j][2];
      }
   }
   cam.target[0]=0.5*(xmin+xmax);
   cam.target[1]=0.5*(ymin+ymax);
   cam.target[2]=0.5*(zmin+zmax);
   cam.dist=sqrt(sqr(xmax-xmin)+sqr(ymax-ymin)+sqr(zmax-zmin));
}

struct CameraButton : public Gluvi::Button {
   CameraButton(const char *t) : Gluvi::Button(t) {}
   void action()
   {
      std::ofstream camera_output("camera.rib");
      Gluvi::camera->export_rib(camera_output);
   }
};

struct ScreenShotButton : public Gluvi::Button{
   const char *filename_format;
   ScreenShotButton(const char *label, const char *filename_format_)
      : Gluvi::Button(label), filename_format(filename_format_)
   {}

   void action()
   { Gluvi::ppm_screenshot(filename_format, frame); }
};

void special_key_handler(int key, int x, int y)
{
   switch(key){
      case GLUT_KEY_LEFT:
         if(read_frame(frame-1))
            glutPostRedisplay();
         break;
      case GLUT_KEY_RIGHT:
         if(read_frame(frame+1))
            glutPostRedisplay();
         break;
      default:
         ;
   }
}

int main(int argc, char **argv)
{
   Gluvi::init(argv[0], &argc, argv);
   Gluvi::Target3D cam;
   Gluvi::camera=&cam;

   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

   Gluvi::userDisplayFunc=display;
   glutSpecialFunc(special_key_handler);
   Gluvi::StaticText frametext_on_screen(frametext);
   Gluvi::root.list.push_back(&frametext_on_screen);
   Gluvi::StaticText keytext("use left/right arrows for frame control");
   Gluvi::root.list.push_back(&keytext);
   CameraButton cambutton("Export camera to RIB");
   Gluvi::root.list.push_back(&cambutton);
   ScreenShotButton screenshot("Screenshot", "frame%04d.ppm");
   Gluvi::root.list.push_back(&screenshot);

   frame=0;
   filename_formats=argv+1;
   mesh.resize(argc-1);
   x.resize(argc-1);
   normal.resize(argc-1);
  
   if(!read_frame(frame)){
      std::cerr<<"Couldn't read first frame"<<std::endl;
      return 1;
   }
   set_camera(cam);

   Gluvi::run();
   return 0;
}

